/* Copyright (C) 2005, 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package compbase;

import osio.TextPrinter;
import debug.DebugWriter;

/**
 * Unit: basic abstraction for a class or interface 
 * 
 * @author S. Frenz
 * @version 090306 added constant for name of static class init method
 *  version 090218 added isOuter method
 *  version 090207 added copyright notice
 *  version 080707 changed signature of searchUnitInView to avoid invalid recursion
 *  version 080706 removed printExts
 *  version 080702 adopted changed symInfo-debug-interface
 *  version 080614 added indirMthdTableEntries, added searching of variables in implemented interfaces
 *  version 080105 added plausibility check in isParent
 *  version 080102 added method to print name of unit including outer units
 *  version 071227 added support for inner units, moved searchCalledUnits to ExAbsCall
 *  version 071215 added enterInheritableReferences to support semantically correct re-filling of class descriptors
 *  version 070816 renamed parameter "name" in searchCalledMethod to clarify difference to Unit.name
 *  version 070801 got offset* from Clss
 *  version 070731 changed extsImplID[] to extsImplIDList
 *  version 070727 got extsID* and implID from Clss and Intf, added symHint
 *  version 070714 added printExts
 *  version 070713 optimized alloceDescriptor
 *  version 070628 adopted changes for "Expr extends TypeRef"
 *  version 070615 took over OutputObject values
 *  version 070527 added final-flag and short-path to inline array variable
 *  version 070522 added flag to support check of descriptor necessity
 *  version 070511 added default method to create constant objects
 *  version 070506 added support for runtime environment without objects
 *  version 070504 added genDescriptor
 *  version 070501 moved prepareOutput hereto
 *  version 070303 added instIndirScalarTableSize;
 *  version 061202 added stat*Table*
 *  version 061030 changed indirectCalls to modifier
 *  version 060723 added flag for explicit standard constructor
 *  version 060607 initial version
 */

public abstract class Unit extends Token {
  public final static int M_FIN    = Modifier.M_FIN;
  public final static int M_PUB    = Modifier.M_PUB;
  public final static int M_STAT   = Modifier.M_STAT;
  public final static int M_PROT   = Modifier.M_PROT;
  public final static int M_PRIV   = Modifier.M_PRIV;
  public final static int M_ABSTR  = Modifier.M_ABSTR;
  public final static int M_INDIR  = Modifier.M_INDIR;
  public final static int M_STRUCT = Modifier.M_STRUCT;
  public final static int MA_ACCSSD = Modifier.MA_ACCSSD;
  
  public final static String OUTERVARNAME = "$outer";
  public final static String STATICMTHDNAME = "$statinit";
  
  //required fields for resolving
  public QualID pack, extsID; //containing package, extended direct unit
  public QualIDList extsImplIDList; //extended/implemented indirect units
  public String name;
  public int modifier, outputAddr;
  public boolean explConstr, explStdConstr; //needs or has explicit constructors, has expl std-constr
  public Vrbl vars, inlArr;
  public Mthd initStat, initDyna, mthds;
  public Unit nextUnit, outerUnit, innerUnits;
  public boolean offsetsAssigned, offsetError, outputGenerated, outputError;
  //required fields for interface-checks, overloading and code-generation
  public int statRelocTableEntries, statScalarTableSize;
  public int clssRelocTableEntries, clssScalarTableSize;
  public int instRelocTableEntries, instScalarTableSize, instIndirScalarTableSize;
  public int indirMthdTableEntries;
  public int symHint; //required for symbolinformer

  private TypeRef typeRefOfThis;

  public Unit(int fid, int il, int ic) {
    super(fid, il, ic);
  }

  //this is what different languages have to implement differently
  public abstract UnitList getRefUnit(Unit refUnit, boolean insert);
  public abstract boolean resolveInterface(Context ctx);
  public abstract boolean hasValidInterface();
  public abstract boolean resolveMethodBlocks(Context ctx);
  public abstract Unit searchUnitInView(StringList name, boolean isRecursion);
  public abstract boolean assignOffsets(boolean doClssOff, Context ctx);
  public abstract boolean genDescriptor(Context ctx);
  public abstract boolean genConstObj(Context ctx);
  public abstract IndirUnitMapList enterInheritableReferences(int objAddr, IndirUnitMapList lastIntf, Context ctx);
  public abstract boolean genOutput(Context ctx);
  public abstract void writeDebug(Context ctx, DebugWriter dbw);

  public TypeRef getTypeTo() {
    if (typeRefOfThis==null) {
      typeRefOfThis=new TypeRef(fileID, line, col);
      typeRefOfThis.baseType=TypeRef.T_QID;
      typeRefOfThis.qid=new QualID(new StringList(name), QualID.Q_UNIT, fileID, line, col);
      typeRefOfThis.qid.unitDest=this;
    }
    return typeRefOfThis;
  }

  public boolean isParent(Unit cmp, Context ctx) {
    Unit u;
    QualIDList list;
    
    //check validity
    if (cmp==null) return false;
    //check ourself
    if (cmp==this) return true;
    //check direct parents
    u=cmp;
    while (u!=null) {
      if (u==this) return true;
      if (u.extsID!=null) {
        if ((u=u.extsID.unitDest)==null) { //destination is not valid ==> resolving error
          compErr(ctx, "unitDest not valid in Unit.isParent");
          ctx.err=true;
          return false;
        }
      }
      else u=null;
    }
    //check indirect parents
    list=cmp.extsImplIDList;
    while (list!=null) {
      if (isParent(list.qid.unitDest, ctx)) return true;
      list=list.nextQualID;
    }
    //not parent
    return false;
  }
  
  public boolean isOuter(Unit inner) {
    while (inner!=null) {
      if (inner==this) return true;
      inner=inner.outerUnit;
    }
    return false;
  }
  
  public Vrbl searchVariable(String cmp, Context ctx) {
    Vrbl me=vars, res=null;
    QualIDList extsList;
    
    while (me!=null) {
      if (me.name.equals(cmp)) return me;
      me=me.nextVrbl;
    }
    if (extsID!=null) {
      if (extsID.unitDest.hasValidInterface()
          && (me=extsID.unitDest.searchVariable(cmp, ctx))!=null) return me;
    }
    extsList=extsImplIDList;
    while (extsList!=null) {
      if (extsList.qid.unitDest.hasValidInterface()
          && (me=extsList.qid.unitDest.searchVariable(cmp, ctx))!=null) {
        if (res==null) res=me;
        else if (res!=me) {
          ctx.out.print("access to variable ");
          ctx.out.print(cmp);
          ctx.out.println(" is ambiguous");
        }
      }
      extsList=extsList.nextQualID;
    }
    return res;
  }
  
  public Mthd searchMethod(Mthd cmp, Context ctx) {
    Mthd me=mthds;
    
    while (me!=null) {
      if (me.matches(cmp, ctx)) return me;
      me=me.nextMthd;
    }
    if (extsID==null) return null;
    if (!extsID.unitDest.hasValidInterface()) return null;
    return extsID.unitDest.searchMethod(cmp, ctx);
  }
  
  protected boolean allocateDescriptor(int scalarSize, int relocEntries, int typeAddr, Context ctx) {
    //check if already prepared
    if (outputAddr!=0) return true;
    //allocate space for class-descriptor
    if ((outputAddr=ctx.mem.allocate(scalarSize, 0, relocEntries, ctx.leanRTE ? -1 : typeAddr))<0) {
      ctx.out.println("not enough memory");
      outputError=true;
      return false;
    }
    return true;
  }

  public void printNameWithOuter(TextPrinter v) {
    if (outerUnit!=null) {
      outerUnit.printNameWithOuter(v);
      v.print(".");
    }
    v.print(name);
  }
}
