/* Copyright (C) 2005, 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package frontend.sjava;

import compbase.*;
import backend.Instruction;

/**
 * StEndLoop: statement to end a loop (break/continue)
 * 
 * @author S. Frenz
 * @version 090207 added copyright notice
 *  version 070911 fixed unnamed continue in switch-statements
 *  version 070909 added support for labels
 *  version 070908 optimized signature of Stmt.resolve
 *  version 070114 reduced access level where possible
 *  version 070111 adapted change in printPos and compErr
 *  version 061231 added check for not null dest
 *  version 061229 removed access to firstInstr
 *  version 061203 optimized calls to printPos and compErr
 *  version 060607 initial version
 */

public class StEndLoop extends Stmt {
	protected final static int BRK = 1;
	protected final static int CNT = 2;
	
	protected int endType;
	protected String labelToEnd;
	private StBreakable outer, toEnd;
	
	protected StEndLoop(StBreakable io, String it, int fid, int il, int ic) {
		super(fid, il, ic);
		outer=io;
		labelToEnd=it;
	}
	
	protected boolean resolve(Unit unitContext, Mthd mthdContext, Context ctx) {
	  StBreakable o;
	  StringList ol;
	  boolean found=false;
	  
    o=outer;
	  if (labelToEnd==null) { //no label given
	    while (o instanceof StBlock || (endType==CNT && o instanceof StSwitch)) o=o.outer;
	    toEnd=o;
	  }
	  else { //search label
	    LOOP: while (o!=null) {
	      ol=o.labels;
	      while (ol!=null) {
	        if (labelToEnd.equals(ol.str)) {
	          toEnd=o;
	          found=true;
	          break LOOP;
	        }
	        ol=ol.next;
	      }
	      o=o.outer;
	    }
	    if (!found) {
	      printPos(ctx, "label not found");
	      return false;
	    }
	  }
    if (toEnd instanceof StFor || toEnd instanceof StWhile //for and while may have all end-stmts
        || (endType==BRK && (toEnd instanceof StSwitch  //switch and stBlock may only have break
            || toEnd instanceof StBlock)))
      return true; 
	  //no valid outer block found
	  printPos(ctx, "statement invalid in this context");
	  return false;
	}
	
	protected void genOutput(Context ctx) {
    Instruction dest;
    
    if ((dest=(endType==CNT) ? toEnd.contDest : toEnd.breakDest)==null) {
      compErr(ctx, "StEndLoop with dest==null");
      return;
    }
    ctx.arch.genJmp(dest);
	}
}
